/* entries.hpp */
/*
Copyright (C) 2010, Igor Yudincev
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the authors may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <string>
#include <unordered_map>


class Entry {
public:
	std::string revision;
	std::string hash;
	Entry() : revision(), hash() {}
	Entry(std::string revision_, std::string hash_)
		: revision(revision_), hash(hash_) {}
	Entry& operator=(const Entry &r) {
		if (this!=&r) {
			revision = r.revision;
			hash = r.hash;
		}
		return *this;
	}
};

class EntriesFile {
private:
	static const int NS100_TIMEOUT = 10000000UL; //   1   

	const wchar_t *dir;
	size_t nDir;
	wchar_t *path;
	bool bPathExists;
	LARGE_INTEGER tChecked; //    /
	FILETIME ftModified;    //  -  
	size_t size;
	char *data;
	char *pExtract;
	std::unordered_map<std::string,Entry> items;
	std::string sDirRevision;

	void invalidate()
	{
		SetLength(&path, 0);
	}

	enum ExtractStringResult {END,OK,PAGE};

	enum ExtractStringResult ExtractString(char **pString, size_t *nString)
    {
    	bool bPrevIsPage = false;
    	*pString = pExtract;
    	*nString = 0;
    	while (pExtract<data+size) {
    		char c = *pExtract++;
    		if (c==0x0A) {
    			if (bPrevIsPage) return PAGE;
    			return OK;
    		}
    		bPrevIsPage = (c==0x0C);
    		(*nString)++;
    	}
    	return END;
    }

	bool parseDir()
    {
		char *pType, *pRevision;
		size_t nType, nRevision;
		enum ExtractStringResult res;

    	pExtract = data;
    	for (int i=0; i<3; i++) {
    		res = ExtractString(&pType,&nType);
    		if (res!=OK) return false;
    	}
    	if (nType==3 && strncmp(pType,"dir",3)==0) {
   			for (int i=0; i<8; i++) {
   				res = ExtractString(&pRevision,&nRevision);
   				if (res!=OK) return false;
   			}
    		std::string sRevision(pRevision,nRevision);
    		sDirRevision = sRevision;
    		return true;
    	}
    	return false;
    }

	bool parseAll()
	{
		char *pDummy, *pFilename, *pType, *pRevision, *pHash;
		size_t nDummy, nFilename, nType, nRevision, nHash;
		enum ExtractStringResult res;

		items.clear();
		if (!parseDir()) return false;

		while ((res = ExtractString(&pDummy,&nDummy))==OK);
		if (res==END) return false;

		for (;;) {
			res = ExtractString(&pFilename,&nFilename);
			if (res!=OK || nFilename==0) break;

			res = ExtractString(&pType,&nType);
			if (res!=OK || nType==0) return false;

			if (nType==4 && strncmp(pType,"file",4)==0) {
				for (int i=0; i<6; i++) {
					res = ExtractString(&pHash,&nHash);
					if (res!=OK) break;
				}
				if (res==END) break;
				if (res==OK && nHash==32) {
					for (int i=0; i<2; i++) {
						res = ExtractString(&pRevision,&nRevision);
						if (res!=OK) return false;
					}
					if (nRevision==0) return false;

					std::string sFilename(pFilename,nFilename);
					std::string sHash(pHash,nHash);
					std::string sRevision(pRevision,nRevision);
					Entry entry(sRevision,sHash);
					items[sFilename] = entry;
				}
			}
			else if (nType==3 && strncmp(pType,"dir",3)==0) {
				wchar_t *pPath=NULL;
				//convert filename to UCS-2
				int nFilenameW = MultiByteToWideChar(CP_UTF8, 0, pFilename, nFilename,
					NULL, 0);
				wchar_t *pFilenameW=NULL;
				SetLength(&pFilenameW, nFilenameW);
				if (MultiByteToWideChar(CP_UTF8, 0, pFilename, nFilename,
					pFilenameW, nFilenameW)!=0)
				{
					SetLength(&pPath, nDir+nFilenameW+1);
					wcscpy(pPath, dir);
					wcscpy(&pPath[nDir], pFilenameW);
					pPath[nDir+nFilenameW] = L'\\';
					EntriesFile* dirEntry = new EntriesFile();
					if (dirEntry->read(pPath,true)) {
						std::string sFilename(pFilename,nFilename);
						std::string sHash("d");
						std::string sRevision(dirEntry->sDirRevision);
						Entry entry(sRevision,sHash);
						items[sFilename] = entry;
					}
					delete dirEntry;
				}
				SetLength(&pFilenameW, 0);
				SetLength(&pPath, 0);
			}

			while (res==OK)
			  res = ExtractString(&pDummy,&nDummy);
			if (res==END) return false;
		}
		return true;
	}

public:
	EntriesFile() : path(NULL), bPathExists(false), data(NULL), sDirRevision("")
	{
		memset(&tChecked, 0, sizeof(tChecked));
	}
	~EntriesFile()
	{
		free(data);
		SetLength(&path, 0);
	}

	///  .svn/entries    
	///\return true,   
	bool read(const wchar_t *dir_, bool bDirOnly=false)
	{
		HANDLE h = INVALID_HANDLE_VALUE;

		dir = dir_;
		nDir=wcslen(dir);
		wchar_t *pPath=NULL;
		SetLength(&pPath, nDir+12);
		wcscpy(pPath, dir);
		wcscpy(&pPath[nDir], L".svn\\entries");

		bool bSamePath = StrCmpIW(path, pPath)==0;
		Assign(&path, pPath);
		SetLength(&pPath, 0);

		FILETIME ftNow;
		LARGE_INTEGER now;
		GetSystemTimeAsFileTime(&ftNow);
		memcpy(&now, &ftNow, sizeof(now));
		bool bCacheOutdated = now.QuadPart-tChecked.QuadPart>=NS100_TIMEOUT;

		if (bSamePath && !bCacheOutdated) { //      
			return bPathExists;
		}

		tChecked = now;

		//      -  entries
		h = CreateFile(path, GENERIC_READ,
			FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, 0);
		bPathExists = h!=INVALID_HANDLE_VALUE;
		if (!bPathExists) return false;

		BY_HANDLE_FILE_INFORMATION info;
		if (!GetFileInformationByHandle(h,&info) ||
			info.nFileSizeLow==0 || info.nFileSizeHigh!=0)
		{
			CloseHandle(h);
			invalidate();
			return false;
		}
		if (bSamePath && CompareFileTime(&info.ftLastWriteTime, &ftModified)==0) {
			CloseHandle(h);
			return true; //unchanged
		}
		ftModified = info.ftLastWriteTime;

		size = info.nFileSizeLow;
		data = (char*)realloc(data, (size_t)size);
		if (data==NULL) {
			CloseHandle(h);
			invalidate();
			return false;
		}
		DWORD n;
		if (!ReadFile(h, data, size, &n, NULL) || size!=n) {
			CloseHandle(h);
			invalidate();
			return false;
		}
		CloseHandle(h);
		return bDirOnly ? parseDir() : parseAll();
	}

	const Entry *find(const char *filename) const
	{
		std::unordered_map<std::string,Entry>::const_iterator it =
			items.find(filename);
		if (it==items.end())
			return NULL;
		return &it->second;
	}

};
